/* $Id$

   Single Button Launcher - launch applications using a single hardware
   button.
   Copyright (C) 2001 Fraser McCrossan http://www.joat.ca

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
*/

#include <PalmOS.h>

#include "resource.h" /* resource IDs first */
#include "sbl.h"

UInt32 rom_version;

Boolean on_poser;

sblpreftype prefs; /* my preferences! */

extern UInt16 current_app; /* which app is the current scan position */

Boolean MiscHandler(EventType *e)
{
  switch (e->eType) {

  case frmOpenEvent:
    return false;

  default:
    return false;
  }
}

Boolean AppHandleEvent(EventType *e)
{
  if (e->eType == frmLoadEvent) {
    FormType *form = FrmInitForm(e->data.frmLoad.formID);
    FrmSetActiveForm(form);
    switch (e->data.frmLoad.formID) {

    case MainForm:
      FrmSetEventHandler(form, MainFormHandler);
      return true;

    default:
      FrmCustomAlert(InternalError, "unhandled form", __FUNCTION__, "");
      return false;
    }
    return false;
  }
  else {
    return false;
  }
}

UInt32 PilotMain(UInt16 cmd, void *cmdPBP UNUSED, UInt16 launchFlags UNUSED)
{
  EventType e;
  UInt16 err;
  UInt16 prefssize;
  UInt16 i;
  Int32 timeout;
  UInt32 keystate = KeyCurrentState();

  if (cmd == sysAppLaunchCmdNormalLaunch) {
    UInt32 value;
    if (FtrGet(sysFtrCreator, sysFtrNumROMVersion, &rom_version) != 0)
      rom_version = 0;  /* Set to very low version if FtrGet failed.  */
    on_poser = (FtrGet('pose', 0, &value) == 0);

    /* set defaults before fetching preferences */
    for (i = 0; i < MaxApps; ++i) {
      StrNCopy(prefs.apps[prefs.lastButtonPressed][i], "", dmDBNameLength);
    }
    prefs.lastButtonPressed = NOLASTBUTTON;
    prefssize = sizeof(prefs);
    if (PrefGetAppPreferences(appSbl, 0, &prefs, &prefssize, true)
	== noPreferenceFound) {
      /* if that call failed, nothing we can do about it, so let's assume
	 that it's because this is the first launch, and display the help */
      FrmHelp(About);
    }
    /* decide which hw button is pressed, and thus which set of apps
     to use; note that before this code runs, lastButtonPressed will
     have whatever value is in the preferences, which may be set by a
     hack */
    if (keystate & keyBitHard1) prefs.lastButtonPressed = 0;
    if (keystate & keyBitHard2) prefs.lastButtonPressed = 1;
    if (keystate & keyBitHard3) prefs.lastButtonPressed = 2;
    if (keystate & keyBitHard4) prefs.lastButtonPressed = 3;
    /* check if it now has a valid value; if it *does*, then just pick some
       arbitrary value */
    if (prefs.lastButtonPressed == NOLASTBUTTON)
      prefs.lastButtonPressed = 3;
    /* find the first assigned button */
    FindFirstButton();
    FrmGotoForm(MainForm);
    do {
      /* timeout below is very low when we are still cycling through the
	 apps, and can return to infinite when we've exhausted them */
      timeout = (Int32) ((current_app < MaxApps) ? (DEFAULTDELAY >> 3) : evtWaitForever);
      EvtGetEvent(&e, timeout);
      if (!MenuHandleEvent(NULL, &e, &err)) {
	if (!AppHandleEvent(&e)) {
	  if (!FrmDispatchEvent(&e)) {
	    /* this is left until last so we can override app buttons */
	    SysHandleEvent(&e);
	  }
	}
      }
    } while (e.eType != appStopEvent);

    FrmCloseAllForms();

    /* make sure the lastButtonPressed is cleared out */
    prefs.lastButtonPressed = NOLASTBUTTON;
    PrefSetAppPreferences(appSbl, 0, 0, (void *) &prefs, sizeof(prefs),
			  true);
  }
  
  return 0;
}
