/*
 * $Id: topalmtext.c,v 1.23 2004/01/09 22:10:50 arpruss Exp $
 *
 * The true home of this program is palmfontconv.sourceforge.net.  If you
 * find it elsewhere and improve it, please send a patch to
 * Alexander R. Pruss <ap85@georgetown.edu>.
 *
 * palmfontconv - Copyright (c) 2003, Alexander R. Pruss
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */


/* Needs FreeType 2 library. */

#include <ctype.h>
#include <stdio.h>
#include <ft2build.h>
#include <freetype/ftglyph.h>
#include FT_FREETYPE_H
#include "encodings.c"

typedef unsigned char Boolean;
typedef unsigned char Byte;

const char grayMap[17] =  "-123456789ABCDE#";
#define MONO_1 ( grayMap[15] )
#define MONO_0 ( grayMap[0] )
unsigned encodingNum;
Boolean  disableKerning = 0;

void SetEncoding( char* name )
{
    int i;
	for ( i = sizeof encodings / sizeof *encodings - 1 ; 0 <= i ; i-- ) {
	     if ( !stricmp( encodings[ i ].name, name ) )
	         break;
	}
    if ( i < 0 )
         i = NULL_ENCODING;
    encodingNum = i;
}



unsigned Decode( unsigned c )
{
    unsigned i;
    for ( i = 0 ; i < encodings[ encodingNum ].numEntries ; i++ ) {
         if ( encodings[ encodingNum ].table[ i ].src == c ) {
             if ( encodings[ encodingNum ].table[ i ].unicode == UNDEFINED_CHAR )
                 return c;
             else
                 return encodings[ encodingNum ].table[ i ].unicode;
         }
    }
    return c;
}



Byte GetRowPixel( Byte* row, int pixel, Boolean gray )
{
    if ( gray )
        return row[ pixel ] / 16;
    else
        return ( row[ pixel / 8 ] >> ( 7 - ( pixel & 7 ) ) ) & 1;
}


void DumpRow( int left0, int right0, Byte* row, int width0, int leftSkip0, Boolean gray, int multiplicity )
{
    int pixel;
    int left;
    int right;
    int width;
    int leftSkip;

    while( 0 < multiplicity-- ) {
        left     = left0;
        right    = right0;
        width    = width0;
        leftSkip = leftSkip0;

        while( 0 < left-- )
            putchar( gray ? grayMap[ 0 ] : MONO_0 );
        pixel = 0;
        while( 0 < width-- ) {
            if ( leftSkip <= 0 ) {
                Byte value;
                value = GetRowPixel( row, pixel, gray );
                if ( ! gray )
                    putchar( value ? MONO_1 : MONO_0 );
                else
                    putchar( grayMap[ value ] );
            }
            else
                leftSkip--;
            pixel++;
        }
        while( 0 < right-- )
            putchar( MONO_0 );
    }
    putchar('\n');
}


int EmitBitmap( int n, FT_Bitmap* bitmap, int left, int xadvance, int top, int height, int multiplicity,
    Boolean getWidth )
{
    int            i;
    int            j;
    int            y;
    unsigned char* buffer;
    int            right;
    Boolean        gray;
    int            blankColsLeft;
    int            leftSkip;
    int            leftKerning;
    int            advance;
    int            noKerning;

    gray = ( bitmap->pixel_mode == ft_pixel_mode_grays );

    if ( 1 < multiplicity )
        noKerning = 1;
    else
        noKerning = disableKerning;

    if ( ! getWidth )
        printf("GLYPH %d\n// left %d, top %d, height %d, xadvance %d, width %d\n",
            n,left,top,height,xadvance,bitmap->width);
    y = 0;
    if ( left < 0 ) {
        blankColsLeft = bitmap->width;
        if ( bitmap->pitch < 0 )
            buffer = bitmap->buffer + ( bitmap->rows - 1 ) * bitmap->pitch;
        else
            buffer = bitmap->buffer;
        for ( i = 0 ; i < bitmap->rows ; i++, buffer += bitmap->pitch ) {
            for ( j = 0 ; j < bitmap->width ; j++ )
                if ( 0 != GetRowPixel( buffer, j, gray ) )
                    break;
            if ( j < blankColsLeft )
                blankColsLeft = j;
        }
        if ( blankColsLeft < -left ) {
            leftSkip = blankColsLeft;
        }
        else {
            leftSkip = -left;
        }
        left += leftSkip;
    }
    else {
        leftSkip = 0;
    }
    if ( ! noKerning ) {
        if ( left <= 0 ) {
            leftKerning = left;
            advance     = xadvance;
        }
        else {
            leftKerning = 0;
            advance     = xadvance;
        }
        if ( ! getWidth )
            printf( "leftKerning %d\n"
                "advance     %d\n",
                leftKerning,
                advance );
    }
    if ( left < 0 ) {
        xadvance -= left;
        left      = 0;
    }
    if ( xadvance < left + bitmap->width - leftSkip ) {
        xadvance = left + bitmap->width - leftSkip;
        right    = 0;
    }
    else {
        right = xadvance - ( left + bitmap->width );
    }
    if ( getWidth )
        return xadvance * multiplicity;
    printf("//Processed: xadvance = %d, leftSkip = %d, left = %d, right = %d\n",
        xadvance,leftSkip,left,right);
    for ( i = 0 ; i < top; i++ ) {
        DumpRow( xadvance, 0, NULL, 0, 0, gray, multiplicity );
        y++;
    }
    if ( bitmap->pitch < 0 )
        buffer = bitmap->buffer + ( bitmap->rows - 1 ) * bitmap->pitch;
    else
        buffer = bitmap->buffer;
    for ( i = 0 ; i < bitmap->rows ; i++, buffer += bitmap->pitch ) {
        DumpRow( left, right, buffer, bitmap->width, leftSkip, gray, multiplicity );
        y++;
    }
    while ( y < height ) {
        DumpRow( xadvance, 0, NULL, 0, 0, gray, multiplicity );
        y++;
    }
    putchar('\n');
    return xadvance * 3;
}


int main( int argc, char** argv )
{
    FT_Library  library;   /* handle to library     */
    FT_Face     face;      /* handle to face object */
    int         i;
    int         pixelSize;
    int         highestTop;
    int         lowestBottom;
    int         argIndex;
    Boolean     gray;
    Boolean     scalable;
    Boolean     forceL1Chars;
    int         forceEven;
    char*       encoding = "palmLatin";

    if ( argc < 2 ) {
        fprintf(stderr,"Usage:\n  %s [-f] [-e] [-encoding] fontname [pixelSize [gray]]\n"
                       "   -f : force 0xA0 to be a non-breaking space and 0x85 to be ellipsis.\n"
                       "   -e : make an attempt to force even width output.\n"
                       "   -encoding : set the encoding if the input font is unicode.\n",argv[0]);
        fprintf(stderr,"The following encodings are supported:\n");
        for ( i = 0 ; i < NUM_ENCODINGS ; i++ )
             fprintf(stderr," %s",encodings[i].name);
        fprintf(stderr,"\n");
        return 5;
    }

    if ( FT_Init_FreeType( &library ) ) {
        fprintf(stderr,"Cannot initialize FreeType library.\n");
        return 4;
    }

    forceEven = 0;

    forceL1Chars = 0;

    argIndex = 1;

    while ( argIndex < argc - 1 && argv[ argIndex ][ 0 ] == '-' ) {
        if ( !strcmp( argv[ argIndex ], "-f" ) ) {
            forceL1Chars = 1;
        }
        else if ( !strcmp( argv[ argIndex ], "-e" ) ) {
            forceEven = 1;
        }
        else if ( !strcmp( argv[ argIndex ], "-e-" ) ) {
            forceEven = 0;
        }
        else if ( !strcmp( argv[ argIndex ], "-n" ) ) {
            disableKerning = 1;
        }
        else if ( !strcmp( argv[ argIndex ], "-f-" ) ) {
            forceL1Chars = 0;
        }
        else if ( !strcmp( argv[ argIndex ], "-n-" ) ) {
            disableKerning = 0;
        }
        else {
            encoding = argv[ argIndex ]+1;
        }
        argIndex++;
    }

    if ( FT_New_Face( library,
                       argv[ argIndex ],
                       0,
                       &face ) ) {
        fprintf(stderr,"Cannot open font file %s.\n", argv[ argIndex ]);
        return 3;
    }

    scalable = FT_IS_SCALABLE( face );
    gray     = ( argc - argIndex == 3 ) && tolower(argv[ argc - 1 ][0])=='g';
    
    if ( ! gray )
        disableKerning = 1;

    if ( ! disableKerning )
        forceEven = 0;

    pixelSize = 22;
    if ( scalable ) {
        if ( 3 <= argc )
            pixelSize = atoi( argv[ argIndex + 1 ] );
        FT_Set_Pixel_Sizes( face, 0, pixelSize );
    }

    if ( FT_Select_Charmap( face, ft_encoding_unicode ) ) {
        SetEncoding( "none" );
        if ( FT_Select_Charmap( face, ft_encoding_latin_1 ) ) {
            if ( face->num_charmaps == 0 )
                printf("// No charmap!\n");
            else {
                printf("// Using first charmap of %d.\n", face->num_charmaps);
                FT_Set_Charmap( face, face->charmaps[ 1 ] );
            }
        }
        else
            printf("// Setting encoding to 8859-1\n");
    }
    else {
        SetEncoding( encoding );
        printf("// Setting encoding to %s\n", encoding);
    }

    printf( "// Converted by %s from %s\n\n", argv[ 0 ], argv[ argIndex ] );

    if ( gray )
        printf( "grayMap %s\nfontVersion 2\n", grayMap );
    else
        printf( "fontType %d\n", 0x9000 );

    highestTop   = 0;
    lowestBottom = 0;
    for ( i = 0 ; i < 256 ; i++ ) {
        /* load glyph image into the slot (erase previous one) */
        int c;
        if ( forceL1Chars && i == 0xA0 )
            c = ' ';
        else if ( forceL1Chars && i == 0x85 )
            c = '.';
        else
            c = i;
        if ( 0 == FT_Get_Char_Index( face, Decode( c ) ) ||
             FT_Load_Char( face, Decode( c ),
                    FT_LOAD_RENDER | ( gray ? FT_RENDER_MODE_NORMAL : FT_LOAD_TARGET_MONO ) ) )
            continue;
        if ( highestTop < face->glyph->bitmap_top ) {
            highestTop = face->glyph->bitmap_top;
        }
        if ( face->glyph->bitmap_top -  face->glyph->bitmap.rows < lowestBottom ) {
            lowestBottom = face->glyph->bitmap_top - face->glyph->bitmap.rows;
        }
    }

    printf( "fRectHeight %d\n", highestTop - lowestBottom );
    printf( "descent %d\n", -lowestBottom );
    printf( "ascent %d\n\n", highestTop );

    for ( i = 0; i < 256; i++ ) {
        int c;
        int multiplicity;

        multiplicity = 1;
        if ( forceL1Chars && i == 0xA0 )
            c = ' ';
        else if ( forceL1Chars && i == 0x85 ) {
            c            = '.';
            multiplicity = 3;
        }
        else
            c = i;
        /* load glyph image into the slot (erase previous one) */
        if ( forceEven && scalable )
            FT_Set_Pixel_Sizes( face, 0, pixelSize );
        if ( FT_Load_Char( face, Decode( c ), FT_LOAD_RENDER | ( gray ? FT_RENDER_MODE_NORMAL : FT_LOAD_TARGET_MONO ) ) )
            continue;
        if ( face->glyph->bitmap.width == 0 && face->glyph->advance.x == 0 )
            continue;
        if ( forceEven && scalable ) {
            int width;
            width = EmitBitmap( i, &(face->glyph->bitmap), face->glyph->bitmap_left,
                ( face->glyph->advance.x + 63 ) / 64,
                highestTop - face->glyph->bitmap_top,
                highestTop - lowestBottom, multiplicity, 1 );
            if ( width % 2 ) {
                int oldWidth;
                /* readjust due to odd width */
                int delta;
                int sign;
                oldWidth = width;
                if ( 32 <= face->glyph->advance.x % 64 || face->glyph->advance.x % 64 == 0 ||
                    ( face->glyph->advance.x + 63 ) / 64 < 5 )
                    sign = 1;
                else
                    sign = -1;
                delta = ( ( pixelSize + width / 2 ) / width ) * sign;
                if ( delta == 0 )
                    delta = sign;
                printf( "// Even width fix from original width %d by using delta=%d.\n", width, delta );
                FT_Set_Pixel_Sizes( face, pixelSize + delta, pixelSize );
                FT_Load_Char( face, Decode( c ), FT_LOAD_RENDER | ( gray ? FT_RENDER_MODE_NORMAL : FT_LOAD_TARGET_MONO ) );
                width = EmitBitmap( i, &(face->glyph->bitmap), face->glyph->bitmap_left,
                  ( face->glyph->advance.x + 63 ) / 64,
                  highestTop - face->glyph->bitmap_top,
                  highestTop - lowestBottom, multiplicity, 1 );
                if ( width % 2 || 1 < width - oldWidth || 1 < oldWidth - width ) {
                    if ( 1 < width - oldWidth && 1 < delta )
                        delta--;
                    else if ( 1 < oldWidth - width && delta < -1 )
                        delta++;
                    else
                        delta = -delta;
                    printf( "// Fix failed; it generated width %d.  Trying delta=%d.\n", width, delta );
                    if ( delta < 0 && ( face->glyph->advance.x + 63 ) / 64 < 5 ) {
                        FT_Set_Pixel_Sizes( face, 0, pixelSize );
                        FT_Load_Char( face, Decode( c ), FT_LOAD_RENDER | ( gray ? FT_RENDER_MODE_NORMAL : FT_LOAD_TARGET_MONO ) );
                    }
                    else {
                        FT_Set_Pixel_Sizes( face, pixelSize + delta, pixelSize );
                        FT_Load_Char( face, Decode( c ), FT_LOAD_RENDER | ( gray ? FT_RENDER_MODE_NORMAL : FT_LOAD_TARGET_MONO ) );
                        width = EmitBitmap( i, &(face->glyph->bitmap), face->glyph->bitmap_left,
                          ( face->glyph->advance.x + 63 ) / 64,
                          highestTop - face->glyph->bitmap_top,
                          highestTop - lowestBottom, multiplicity, 1 );
                        if ( width % 2 || 1 < width - oldWidth || 1 < oldWidth - width ) {
                            printf( "// Second fix attempt failed, too, generating width %d.\n", width );
                            FT_Set_Pixel_Sizes( face, 0, pixelSize );
                        }
                    }
                }
            }
        }
        /* now, draw to our target surface */
        EmitBitmap( i, &(face->glyph->bitmap), face->glyph->bitmap_left,
              ( face->glyph->advance.x + 63 ) / 64,
              highestTop - face->glyph->bitmap_top,
              highestTop - lowestBottom, multiplicity, 0 );
      }
     return 0;
}
